@@ -2,9 +2,22 @@
 /* maintains the internal dive list structure */
 #include <string.h>
 #include <stdio.h>
-#include <glib/gi18n.h>
-
+#include <limits.h>
+#include "gettext.h"
 #include "dive.h"
+#include "planner.h"
+
+struct tag_entry *g_tag_list = NULL;
+
+static const char* default_tags[] = {
+	QT_TRANSLATE_NOOP("gettextFromC", "boat"), QT_TRANSLATE_NOOP("gettextFromC", "shore"), QT_TRANSLATE_NOOP("gettextFromC", "drift"),
+	QT_TRANSLATE_NOOP("gettextFromC", "deep"), QT_TRANSLATE_NOOP("gettextFromC", "cavern") , QT_TRANSLATE_NOOP("gettextFromC", "ice"),
+	QT_TRANSLATE_NOOP("gettextFromC", "wreck"), QT_TRANSLATE_NOOP("gettextFromC", "cave"), QT_TRANSLATE_NOOP("gettextFromC", "altitude"),
+	QT_TRANSLATE_NOOP("gettextFromC", "pool"), QT_TRANSLATE_NOOP("gettextFromC", "lake"), QT_TRANSLATE_NOOP("gettextFromC", "river"),
+	QT_TRANSLATE_NOOP("gettextFromC", "night"), QT_TRANSLATE_NOOP("gettextFromC", "fresh"), QT_TRANSLATE_NOOP("gettextFromC", "student"),
+	QT_TRANSLATE_NOOP("gettextFromC", "instructor"), QT_TRANSLATE_NOOP("gettextFromC", "photo"), QT_TRANSLATE_NOOP("gettextFromC", "video"),
+	QT_TRANSLATE_NOOP("gettextFromC", "deco")
+};
 
 void add_event(struct divecomputer *dc, int time, int type, int flags, int value, const char *name)
 {
@@ -41,15 +54,15 @@ int get_pressure_units(unsigned int mb, const char **units)
 	switch (units_p->pressure) {
 	case PASCAL:
 		pressure = mb * 100;
-		unit = _("pascal");
+		unit = translate("gettextFromC","pascal");
 		break;
 	case BAR:
 		pressure = (mb + 500) / 1000;
-		unit = _("bar");
+		unit = translate("gettextFromC","bar");
 		break;
 	case PSI:
 		pressure = mbar_to_PSI(mb);
-		unit = _("psi");
+		unit = translate("gettextFromC","psi");
 		break;
 	}
 	if (units)
@@ -85,12 +98,12 @@ double get_volume_units(unsigned int ml, int *frac, const char **units)
 	switch (units_p->volume) {
 	case LITER:
 		vol = ml / 1000.0;
-		unit = _("l");
+		unit = translate("gettextFromC","l");
 		decimals = 1;
 		break;
 	case CUFT:
 		vol = ml_to_cuft(ml);
-		unit = _("cuft");
+		unit = translate("gettextFromC","cuft");
 		decimals = 2;
 		break;
 	}
@@ -101,6 +114,13 @@ double get_volume_units(unsigned int ml, int *frac, const char **units)
 	return vol;
 }
 
+unsigned int units_to_depth(double depth)
+{
+	if (get_units()->length == METERS)
+		return rint(depth * 1000);
+	return feet_to_mm(depth);
+}
+
 double get_depth_units(unsigned int mm, int *frac, const char **units)
 {
 	int decimals;
@@ -111,12 +131,12 @@ double get_depth_units(unsigned int mm, int *frac, const char **units)
 	switch (units_p->length) {
 	case METERS:
 		d = mm / 1000.0;
-		unit = _("m");
+		unit = translate("gettextFromC","m");
 		decimals = d < 20;
 		break;
 	case FEET:
 		d = mm_to_feet(mm);
-		unit = _("ft");
+		unit = translate("gettextFromC","ft");
 		decimals = 0;
 		break;
 	}
@@ -127,6 +147,30 @@ double get_depth_units(unsigned int mm, int *frac, const char **units)
 	return d;
 }
 
+double get_vertical_speed_units(unsigned int mms, int *frac, const char **units)
+{
+	double d;
+	const char *unit;
+	const struct units *units_p = get_units();
+	const double time_factor = units_p->vertical_speed_time == MINUTES ? 60.0 : 1.0;
+
+	switch (units_p->length) {
+	case METERS:
+		d = mms / 1000.0 * time_factor;
+		unit = translate("gettextFromC",(units_p->vertical_speed_time == MINUTES) ? "m/min" : "m/s");
+		break;
+	case FEET:
+		d = mm_to_feet(mms) * time_factor;
+		unit = translate("gettextFromC",(units_p->vertical_speed_time == MINUTES) ? "ft/min" : "ft/s");
+		break;
+	}
+	if (frac)
+		*frac = d < 10;
+	if (units)
+		*units = unit;
+	return d;
+}
+
 double get_weight_units(unsigned int grams, int *frac, const char **units)
 {
 	int decimals;
@@ -136,11 +180,11 @@ double get_weight_units(unsigned int grams, int *frac, const char **units)
 
 	if (units_p->weight == LBS) {
 		value = grams_to_lbs(grams);
-		unit = _("lbs");
+		unit = translate("gettextFromC","lbs");
 		decimals = 0;
 	} else {
 		value = grams / 1000.0;
-		unit = _("kg");
+		unit = translate("gettextFromC","kg");
 		decimals = 1;
 	}
 	if (frac)
@@ -158,9 +202,46 @@ struct dive *alloc_dive(void)
 	if (!dive)
 		exit(1);
 	memset(dive, 0, sizeof(*dive));
+	taglist_init(&(dive->tag_list));
 	return dive;
 }
 
+/* only copies events from the first dive computer */
+void copy_events(struct dive *s, struct dive *d)
+{
+	struct event *ev;
+	if (!s || !d)
+		return;
+	ev = s->dc.events;
+	d->dc.events = NULL;
+	while (ev != NULL) {
+		add_event(&d->dc, ev->time.seconds, ev->type, ev->flags, ev->value, ev->name);
+		ev = ev->next;
+	}
+}
+
+void copy_cylinders(struct dive *s, struct dive *d)
+{
+	int i;
+	if (!s || !d)
+		return;
+	for (i = 0; i < MAX_CYLINDERS; i++)
+		d->cylinder[i] = s->cylinder[i];
+}
+
+void copy_samples(struct dive *s, struct dive *d)
+{
+	/* instead of carefully copying them one by one and calling add_sample
+	 * over and over again, let's just copy the whole blob */
+	if (!s || !d)
+		return;
+	int nr = s->dc.samples;
+	d->dc.samples = nr;
+	d->dc.sample = malloc(nr * sizeof(struct sample));
+	if (d->dc.sample)
+		memcpy(d->dc.sample, s->dc.sample, nr * sizeof(struct sample));
+}
+
 struct sample *prepare_sample(struct divecomputer *dc)
 {
 	if (dc) {
@@ -259,6 +340,44 @@ static void fixup_dc_duration(struct divecomputer *dc)
 	}
 }
 
+void per_cylinder_mean_depth(struct dive *dive, struct divecomputer *dc, int *mean, int *duration)
+{
+	int i;
+	int depthtime[MAX_CYLINDERS] = {0,};
+	int lasttime = 0, lastdepth = 0;
+	int idx = 0;
+
+	for (i = 0; i < MAX_CYLINDERS; i++)
+		mean[i] = duration[i] = 0;
+	struct event *ev = get_next_event(dc->events, "gaschange");
+	if (!ev) {
+		// special case - no gas changes
+		mean[0] = dc->meandepth.mm;
+		duration[0] = dc->duration.seconds;
+		return;
+	}
+	for (i = 0; i < dc->samples; i++) {
+		struct sample *sample = dc->sample + i;
+		int time = sample->time.seconds;
+		int depth = sample->depth.mm;
+		if (ev && time >= ev->time.seconds) {
+			idx = get_cylinder_index(dive, ev);
+			ev = get_next_event(ev->next, "gaschange");
+		}
+		/* We ignore segments at the surface */
+		if (depth > SURFACE_THRESHOLD || lastdepth > SURFACE_THRESHOLD) {
+			duration[idx] += time - lasttime;
+			depthtime[idx] += (time - lasttime) * (depth + lastdepth) / 2;
+		}
+		lastdepth = depth;
+		lasttime = time;
+	}
+	for (i = 0; i < MAX_CYLINDERS; i++) {
+		if (duration[i])
+			mean[i] = (depthtime[i] + duration[i] / 2) / duration[i];
+	}
+}
+
 static void fixup_pressure(struct dive *dive, struct sample *sample)
 {
 	unsigned int pressure, index;
@@ -312,6 +431,11 @@ int gas_volume(cylinder_t *cyl, pressure_t p)
 	return cyl->type.size.mliter * surface_volume_multiplier(p);
 }
 
+int wet_volume(double cuft, pressure_t p)
+{
+	return cuft_to_l(cuft) * 1000 / surface_volume_multiplier(p);
+}
+
 /*
  * If the cylinder tank pressures are within half a bar
  * (about 8 PSI) of the sample pressures, we consider it
@@ -439,7 +563,7 @@ static void sanitize_cylinder_info(struct dive *dive)
 }
 
 /* some events should never be thrown away */
-static gboolean is_potentially_redundant(struct event *event)
+static bool is_potentially_redundant(struct event *event)
 {
 	if (!strcmp(event->name, "gaschange"))
 		return FALSE;
@@ -522,19 +646,29 @@ static void fixup_duration(struct dive *dive)
 	dive->duration.seconds = duration;
 }
 
-static void fixup_watertemp(struct dive *dive)
+/*
+ * What do the dive computers say the water temperature is?
+ * (not in the samples, but as dc property for dcs that support that)
+ */
+unsigned int dc_watertemp(struct divecomputer *dc)
 {
-	struct divecomputer *dc;
 	int sum = 0, nr = 0;
 
-	for_each_dc(dive, dc) {
+	do {
 		if (dc->watertemp.mkelvin) {
 			sum += dc->watertemp.mkelvin;
 			nr++;
 		}
-	}
-	if (nr)
-		dive->watertemp.mkelvin = (sum + nr / 2) / nr;
+	} while ((dc = dc->next) != NULL);
+	if (!nr)
+		return 0;
+	return (sum + nr / 2) / nr;
+}
+
+static void fixup_watertemp(struct dive *dive)
+{
+	if (!dive->watertemp.mkelvin)
+		dive->watertemp.mkelvin = dc_watertemp(&dive->dc);
 }
 
 /*
@@ -729,10 +863,6 @@ struct dive *fixup_dive(struct dive *dive)
 	int i;
 	struct divecomputer *dc;
 
-	add_people(dive->buddy);
-	add_people(dive->divemaster);
-	add_location(dive->location);
-	add_suit(dive->suit);
 	sanitize_cylinder_info(dive);
 	dive->maxcns = dive->cns;
 
@@ -918,7 +1048,7 @@ static char *merge_text(const char *a, const char *b)
 	res = malloc(strlen(a) + strlen(b) + 32);
 	if (!res)
 		return (char *)a;
-	sprintf(res, _("(%s) or (%s)"), a, b);
+	sprintf(res, translate("gettextFromC","(%s) or (%s)"), a, b);
 	return res;
 }
 
@@ -1056,20 +1186,13 @@ static int find_cylinder_match(cylinder_t *cyl, cylinder_t array[], unsigned int
 /* Force an initial gaschange event to the (old) gas #0 */
 static void add_initial_gaschange(struct dive *dive, struct divecomputer *dc)
 {
-	int o2, he, value;
 	struct event *ev = get_next_event(dc->events, "gaschange");
 
 	if (ev && ev->time.seconds < 30)
 		return;
 
 	/* Old starting gas mix */
-	o2 = get_o2(&dive->cylinder[0].gasmix);
-	he = get_he(&dive->cylinder[0].gasmix);
-	o2 = (o2 + 5) / 10;
-	he = (he + 5) / 10;
-	value = o2 + (he << 16);
-
-	add_event(dc, 0, 11, 0, value, "gaschange");
+	add_gas_switch_event(dive, dc, 0, 0);
 }
 
 static void dc_cylinder_renumber(struct dive *dive, struct divecomputer *dc, int mapping[])
@@ -1543,7 +1666,7 @@ static int likely_same_dive(struct dive *a, struct dive *b)
  * merges almost exact duplicates - something that happens easily
  * with overlapping dive downloads.
  */
-struct dive *try_to_merge(struct dive *a, struct dive *b, gboolean prefer_downloaded)
+struct dive *try_to_merge(struct dive *a, struct dive *b, bool prefer_downloaded)
 {
 	if (likely_same_dive(a, b))
 		return merge_dives(a, b, 0, prefer_downloaded);
@@ -1768,7 +1891,161 @@ static void join_dive_computers(struct divecomputer *res, struct divecomputer *a
 	remove_redundant_dc(res, prefer_downloaded);
 }
 
-struct dive *merge_dives(struct dive *a, struct dive *b, int offset, gboolean prefer_downloaded)
+int taglist_get_tagstring(struct tag_entry *tag_list, char *buffer, int len) {
+	int i = 0;
+	struct tag_entry *tmp;
+	tmp = tag_list->next;
+	memset(buffer, 0, len);
+	while(tmp != NULL) {
+		int newlength = strlen(tmp->tag->name);
+		if (i > 0)
+			newlength += 2;
+		if ((i+newlength) < len) {
+			if (i > 0) {
+				strcpy(buffer+i, ", ");
+				strcpy(buffer+i+2, tmp->tag->name);
+			} else {
+				strcpy(buffer, tmp->tag->name);
+			}
+		} else {
+			return i;
+		}
+		i += newlength;
+		tmp = tmp->next;
+	}
+	return i;
+}
+
+struct divetag *taglist_get_tag(struct tag_entry *tag_list, const char *tag)
+{
+	struct tag_entry *tmp;
+	tmp = tag_list->next;
+	while(tmp != NULL) {
+		if (tmp->tag != NULL) {
+			if (strcmp(tmp->tag->name, tag) == 0)
+				return tmp->tag;
+			else
+				tmp = tmp->next;
+		}
+	}
+	return NULL;
+}
+
+static inline void taglist_free_divetag(struct divetag *tag) {
+	if (tag->name != NULL)
+		free(tag->name);
+	if (tag->source != NULL)
+		free(tag->source);
+	free(tag);
+}
+
+/* Add a tag to the tag_list, keep the list sorted */
+static struct divetag *taglist_add_divetag(struct tag_entry *tag_list, struct divetag *tag)
+{
+	struct tag_entry *tmp, *last;
+	last = tag_list;
+	tmp = tag_list->next;
+	while(1) {
+		if (tmp == NULL || strcmp(tmp->tag->name, tag->name) > 0) {
+			/* Insert in front of it */
+			last->next = malloc(sizeof(struct tag_entry));
+			last->next->next = tmp;
+			last->next->tag = tag;
+			return last->next->tag;
+		} else if (strcmp(tmp->tag->name, tag->name) == 0) {
+			/* Already in list */
+			return tmp->tag;
+		} else {
+			last = tmp;
+			tmp = tmp->next;
+		}
+	}
+}
+
+struct divetag *taglist_add_tag(struct tag_entry *tag_list, const char *tag)
+{
+	int i = 0, is_default_tag = 0;
+	struct divetag *ret_tag, *new_tag;
+	const char *translation;
+	new_tag = malloc(sizeof(struct divetag));
+
+	for (i=0; i<sizeof(default_tags)/sizeof(char*); i++) {
+		if (strcmp(default_tags[i], tag) == 0) {
+			is_default_tag = 1;
+			break;
+		}
+	}
+	/* Only translate default tags */
+	if (is_default_tag) {
+		translation = translate("gettextFromC", tag);
+		new_tag->name = malloc(strlen(translation)+1);
+		memcpy(new_tag->name, translation, strlen(translation)+1);
+		new_tag->source = malloc(strlen(tag)+1);
+		memcpy(new_tag->source, tag, strlen(tag)+1);
+	} else {
+		new_tag->source = NULL;
+		new_tag->name = malloc(strlen(tag)+1);
+		memcpy(new_tag->name, tag, strlen(tag)+1);
+	}
+	/* Try to insert new_tag into g_tag_list if we are not operating on it */
+	if (tag_list != g_tag_list) {
+		ret_tag = taglist_add_divetag(g_tag_list, new_tag);
+		/* g_tag_list already contains new_tag, free the duplicate */
+		if (ret_tag != new_tag)
+			taglist_free_divetag(new_tag);
+		ret_tag = taglist_add_divetag(tag_list, ret_tag);
+	} else {
+		ret_tag = taglist_add_divetag(tag_list, new_tag);
+		if (ret_tag != new_tag)
+			taglist_free_divetag(new_tag);
+	}
+	return ret_tag;
+}
+
+void taglist_init(struct tag_entry **tag_list) {
+	*tag_list = malloc(sizeof(struct tag_entry));
+	(*tag_list)->next = NULL;
+	(*tag_list)->tag = NULL;
+}
+
+/* Clear everything but the first element */
+void taglist_clear(struct tag_entry *tag_list) {
+	struct tag_entry *current_tag_entry, *next;
+	current_tag_entry = tag_list->next;
+	while (current_tag_entry != NULL) {
+		next = current_tag_entry->next;
+		free(current_tag_entry);
+		current_tag_entry = next;
+	}
+	tag_list->next = NULL;
+}
+
+/* Merge src1 and src2, write to *dst */
+static void taglist_merge(struct tag_entry *dst, struct tag_entry *src1, struct tag_entry *src2)
+{
+	struct tag_entry *current_tag_entry;
+	current_tag_entry = src1->next;
+	while (current_tag_entry != NULL) {
+		taglist_add_divetag(dst, current_tag_entry->tag);
+		current_tag_entry = current_tag_entry->next;
+	}
+	current_tag_entry = src2->next;
+	while (current_tag_entry != NULL) {
+		taglist_add_divetag(dst, current_tag_entry->tag);
+		current_tag_entry = current_tag_entry->next;
+	}
+}
+
+void taglist_init_global()
+{
+	int i;
+	taglist_init(&g_tag_list);
+
+	for(i=0; i<sizeof(default_tags)/sizeof(char*); i++)
+		taglist_add_tag(g_tag_list, default_tags[i]);
+}
+
+struct dive *merge_dives(struct dive *a, struct dive *b, int offset, bool prefer_downloaded)
 {
 	struct dive *res = alloc_dive();
 	struct dive *dl = NULL;
@@ -1796,6 +2073,7 @@ struct dive *merge_dives(struct dive *a, struct dive *b, int offset, gboolean pr
 	MERGE_MAX(res, a, b, number);
 	MERGE_NONZERO(res, a, b, cns);
 	MERGE_NONZERO(res, a, b, visibility);
+	taglist_merge(res->tag_list, a->tag_list, b->tag_list);
 	merge_equipment(res, a, b);
 	merge_airtemps(res, a, b);
 	if (dl) {
@@ -1825,7 +2103,7 @@ struct dive *find_dive_including(timestamp_t when)
 	return NULL;
 }
 
-gboolean dive_within_time_range(struct dive *dive, timestamp_t when, timestamp_t offset)
+bool dive_within_time_range(struct dive *dive, timestamp_t when, timestamp_t offset)
 {
 	return when - offset <= dive->when && dive->when + dive->duration.seconds <= when + offset;
 }
@@ -1844,3 +2122,15 @@ struct dive *find_dive_n_near(timestamp_t when, int n, timestamp_t offset)
 	}
 	return NULL;
 }
+
+void shift_times(const timestamp_t amount)
+{
+	int i;
+	struct dive *dive;
+
+	for_each_dive (i, dive) {
+		if (!dive->selected)
+			continue;
+		dive->when += amount;
+	}
+}
