@@ -11,30 +11,45 @@
 #include <libxml/parserInternals.h>
 #include <libxml/tree.h>
 #include <libxslt/transform.h>
-#include <glib/gi18n.h>
+
+#include "gettext.h"
+
 #include<sqlite3.h>
 
 #include "dive.h"
 #include "device.h"
 
-int verbose;
+int verbose, quit;
 
-static xmlDoc *test_xslt_transforms(xmlDoc *doc, GError **error);
+static xmlDoc *test_xslt_transforms(xmlDoc *doc, const char **params, char **error);
+char *xslt_path;
 
 /* the dive table holds the overall dive list; target table points at
  * the table we are currently filling */
 struct dive_table dive_table;
 struct dive_table *target_table = NULL;
 
-static void parser_error(GError **error, const char *fmt, ...)
+static void parser_error(char **error, const char *fmt, ...)
 {
 	va_list args;
+	char *tmp;
 
 	if (!error)
 		return;
+
+	tmp = malloc(1024);
 	va_start(args, fmt);
-	*error = g_error_new_valist(g_quark_from_string("subsurface"), DIVE_ERROR_PARSE, fmt, args);
+	vsnprintf(tmp, 1024, fmt, args);
 	va_end(args);
+
+	if (*error) {
+		int len = strlen(*error) + strlen(tmp) + 1;
+		*error = realloc(*error, len);
+		strncat(*error, tmp, strlen(tmp));
+		free(tmp);
+	} else {
+		*error = tmp;
+	}
 }
 
 /*
@@ -80,12 +95,16 @@ static void nonmatch(const char *type, const char *name, char *buffer)
 typedef void (*matchfn_t)(char *buffer, void *);
 
 static int match(const char *pattern, int plen,
-		 const char *name, int nlen,
+		 const char *name,
 		 matchfn_t fn, char *buf, void *data)
 {
-	if (plen > nlen)
+	switch (name[plen]) {
+	case '\0': case '.':
+		break;
+	default:
 		return 0;
-	if (memcmp(pattern, name + nlen - plen, plen))
+	}
+	if (memcmp(pattern, name, plen))
 		return 0;
 	fn(buf, data);
 	return 1;
@@ -116,7 +135,7 @@ struct {
 	const char *nickname, *serial_nr, *firmware;
 } dc;
 } cur_settings;
-static gboolean in_settings = FALSE;
+static bool in_settings = FALSE;
 static struct tm cur_tm;
 static int cur_cylinder_index, cur_ws_index;
 static int lastndl, laststoptime, laststopdepth, lastcns, lastpo2, lastindeco;
@@ -195,24 +214,47 @@ static void divedatetime(char *buffer, void *_when)
 	}
 }
 
+enum ParseState {FINDSTART, FINDEND};
 static void divetags(char *buffer, void *_tags)
 {
-	int *tags = _tags;
-	int i;
+	struct tag_entry *tags = _tags;
+	int i = 0, start = 0, end = 0;
+	enum ParseState state = FINDEND;
+	int len = buffer ? strlen(buffer) : 0;
 
-	for (i = 0; i < DTAG_NR; i++) {
-		if (strstr(buffer, dtag_names[i])) {
-			/* stupidly we have 'cave' and 'cavern' */
-			if (1 << i == DTAG_CAVE) {
-				char *cave = strstr(buffer, "cave");
-				while (cave && !strncmp(cave, "cavern", strlen("cavern"))) {
-					cave++;
-					cave = strstr(cave, "cave");
+	while(i < len) {
+		if (buffer[i] == ',') {
+			if (state == FINDSTART) {
+				/* Detect empty tags */
+			} else if (state == FINDEND) {
+				/* Found end of tag */
+				if (i > 0 && buffer[i - 1] != '\\') {
+						buffer[i] = '\0';
+						state=FINDSTART;
+						taglist_add_tag(tags, buffer+start);
+				} else {
+					state=FINDSTART;
 				}
-				if (!cave)
-					continue;
 			}
-			*tags |= (1 << i);
+		} else if (buffer[i] == ' ') {
+			/* Handled */
+		} else {
+			/* Found start of tag */
+			if (state == FINDSTART) {
+				state = FINDEND;
+				start = i;
+			} else if (state == FINDEND) {
+				end = i;
+			}
+		}
+		i++;
+	}
+	if (state == FINDEND) {
+		if (end < start)
+			end = len - 1;
+		if (len > 0) {
+			buffer[end + 1] = '\0';
+			taglist_add_tag(tags, buffer + start);
 		}
 	}
 }
@@ -222,13 +264,109 @@ enum number_type {
 	FLOAT
 };
 
+double ascii_strtod(char *str, char **ptr)
+{
+	char *p = str, c, *ep;
+	double val = 0.0;
+	double decimal;
+	int sign = 0, esign = 0;
+	int numbers = 0, dot = 0;
+
+	/* skip spaces */
+	while (isspace(c = *p++))
+		/* */;
+
+	/* optional sign */
+	switch (c) {
+	case '-':
+		sign = 1;
+		/* fallthrough */
+	case '+':
+		c = *p++;
+	}
+
+	/* Mantissa */
+	for (;;c = *p++) {
+		if (c == '.') {
+			if (dot)
+				goto done;
+			dot = 1;
+			decimal = 1.0;
+			continue;
+		}
+		if (c >= '0' && c <= '9') {
+			numbers++;
+			if (dot) {
+				decimal /= 10;
+				val += (c - '0') * decimal;
+			} else {
+				val = (val * 10) + (c - '0');
+			}
+			continue;
+		}
+		if (c != 'e' && c != 'E')
+			goto done;
+		break;
+	}
+
+	if (!numbers)
+		goto done;
+
+	/* Exponent */
+	ep = p;
+	c = *ep++;
+	switch (c) {
+	case '-':
+		esign = 1;
+		/* fallthrough */
+	case '+':
+		c = *ep++;
+	}
+
+	if (c >= '0' && c <= '9') {
+		p = ep;
+		int exponent = c - '0';
+
+		for (;;) {
+			c = *p++;
+			if (c < '0' || c > '9')
+				break;
+			exponent *= 10;
+			exponent += c - '0';
+		}
+
+		/* We're not going to bother playing games */
+		if (exponent > 308)
+			exponent = 308;
+
+		while (exponent-- > 0) {
+			if (esign)
+				val /= 10;
+			else
+				val *= 10;
+		}
+	}
+
+done:
+	if (!numbers)
+		goto no_conversion;
+	if (ptr)
+		*ptr = p-1;
+	return sign ? -val : val;
+
+no_conversion:
+	if (ptr)
+		*ptr = str;
+	return 0.0;
+}
+
 static enum number_type parse_float(char *buffer, double *res, char **endp)
 {
 	double val;
-	static gboolean first_time = TRUE;
+	static bool first_time = TRUE;
 
 	errno = 0;
-	val = g_ascii_strtod(buffer, endp);
+	val = ascii_strtod(buffer, endp);
 	if (errno || *endp == buffer)
 		return NEITHER;
 	if (**endp == ',') {
@@ -242,7 +380,7 @@ static enum number_type parse_float(char *buffer, double *res, char **endp)
 			}
 			/* Try again */
 			**endp = '.';
-			val = g_ascii_strtod(buffer, endp);
+			val = ascii_strtod(buffer, endp);
 		}
 	}
 
@@ -282,7 +420,7 @@ static void pressure(char *buffer, void *_press)
 				mbar = mbar * 1000;
 			break;
 		case PSI:
-			mbar = val.fp * 68.95;
+			mbar = psi_to_mbar(val.fp);
 			break;
 		}
 		if (mbar > 5 && mbar < 500000) {
@@ -320,7 +458,7 @@ static void depth(char *buffer, void *_depth)
 			depth->mm = val.fp * 1000 + 0.5;
 			break;
 		case FEET:
-			depth->mm = val.fp * 304.8 + 0.5;
+			depth->mm = feet_to_mm(val.fp);
 			break;
 		}
 		break;
@@ -341,7 +479,7 @@ static void weight(char *buffer, void *_weight)
 			weight->grams = val.fp * 1000 + 0.5;
 			break;
 		case LBS:
-			weight->grams = val.fp * 453.6 + 0.5;
+			weight->grams = lbs_to_grams(val.fp);
 			break;
 		}
 		break;
@@ -362,10 +500,10 @@ static void temperature(char *buffer, void *_temperature)
 			temperature->mkelvin = val.fp * 1000;
 			break;
 		case CELSIUS:
-			temperature->mkelvin = val.fp * 1000 + ZERO_C_IN_MKELVIN + 0.5;
+			temperature->mkelvin = C_to_mkelvin(val.fp);
 			break;
 		case FAHRENHEIT:
-			temperature->mkelvin = (val.fp + 459.67) * 5000/9;
+			temperature->mkelvin = F_to_mkelvin(val.fp);
 			break;
 		}
 		break;
@@ -374,7 +512,7 @@ static void temperature(char *buffer, void *_temperature)
 	}
 	/* temperatures outside -40C .. +70C should be ignored */
 	if (temperature->mkelvin < ZERO_C_IN_MKELVIN - 40000 ||
-	    temperature->mkelvin > ZERO_C_IN_MKELVIN + 70000)
+		temperature->mkelvin > ZERO_C_IN_MKELVIN + 70000)
 		temperature->mkelvin = 0;
 }
 
@@ -408,6 +546,7 @@ static void duration(char *buffer, void *_time)
 		char *dot = strchr(mybuffer,'.');
 		*dot = ':';
 		sampletime(mybuffer, _time);
+		free(mybuffer);
 	} else {
 		sampletime(buffer, _time);
 	}
@@ -423,7 +562,7 @@ static void percent(char *buffer, void *_fraction)
 	case FLOAT:
 		/* Turn fractions into percent unless explicit.. */
 		if (val <= 1.0) {
-			while (g_ascii_isspace(*end))
+			while (isspace(*end))
 				end++;
 			if (*end != '%')
 				val *= 100;
@@ -435,7 +574,7 @@ static void percent(char *buffer, void *_fraction)
 			break;
 		}
 	default:
-		printf(_("Strange percentage reading %s\n"), buffer);
+		printf(translate("gettextFromC","Strange percentage reading %s\n"), buffer);
 		break;
 	}
 }
@@ -474,10 +613,10 @@ static void utf8_string(char *buffer, void *_res)
 {
 	int size;
 	char *res;
-	while (g_ascii_isspace(*buffer))
+	while (isspace(*buffer))
 		buffer++;
 	size = strlen(buffer);
-	while (size && g_ascii_isspace(buffer[size-1]))
+	while (size && isspace(buffer[size-1]))
 		size--;
 	if (!size)
 		return;
@@ -488,7 +627,7 @@ static void utf8_string(char *buffer, void *_res)
 }
 
 #define MATCH(pattern, fn, dest) \
-	match(pattern, strlen(pattern), name, len, fn, buf, dest)
+	match(pattern, strlen(pattern), name, fn, buf, dest)
 
 static void get_index(char *buffer, void *_i)
 {
@@ -508,13 +647,13 @@ static void get_rating(char *buffer, void *_i)
 static void double_to_permil(char *buffer, void *_i)
 {
 	int *i = _i;
-	*i = g_ascii_strtod(buffer, NULL) * 1000.0 + 0.5;
+	*i = ascii_strtod(buffer, NULL) * 1000.0 + 0.5;
 }
 
 static void hex_value(char *buffer, void *_i)
 {
 	uint32_t *i = _i;
-	*i = strtol(buffer, NULL, 16);
+	*i = strtoul(buffer, NULL, 16);
 }
 
 static void get_tripflag(char *buffer, void *_tf)
@@ -603,12 +742,12 @@ static void psi_or_bar(char *buffer, void *_pressure)
 	}
 }
 
-static int divinglog_fill_sample(struct sample *sample, const char *name, int len, char *buf)
+static int divinglog_fill_sample(struct sample *sample, const char *name, char *buf)
 {
-	return	MATCH(".p.time", sampletime, &sample->time) ||
-		MATCH(".p.depth", depth, &sample->depth) ||
-		MATCH(".p.temp", fahrenheit, &sample->temperature) ||
-		MATCH(".p.press1", psi_or_bar, &sample->cylinderpressure) ||
+	return	MATCH("time.p", sampletime, &sample->time) ||
+		MATCH("depth.p", depth, &sample->depth) ||
+		MATCH("temp.p", fahrenheit, &sample->temperature) ||
+		MATCH("press1.p", psi_or_bar, &sample->cylinderpressure) ||
 		0;
 }
 
@@ -623,13 +762,13 @@ static void uddf_gasswitch(char *buffer, void *_sample)
 	add_gas_switch_event(dive, dc, seconds, idx);
 }
 
-static int uddf_fill_sample(struct sample *sample, const char *name, int len, char *buf)
+static int uddf_fill_sample(struct sample *sample, const char *name, char *buf)
 {
-	return	MATCH(".divetime", sampletime, &sample->time) ||
-		MATCH(".depth", depth, &sample->depth) ||
-		MATCH(".temperature", temperature, &sample->temperature) ||
-		MATCH(".tankpressure", pressure, &sample->cylinderpressure) ||
-		MATCH(".switchmix.ref", uddf_gasswitch, sample) ||
+	return	MATCH("divetime", sampletime, &sample->time) ||
+		MATCH("depth", depth, &sample->depth) ||
+		MATCH("temperature", temperature, &sample->temperature) ||
+		MATCH("tankpressure", pressure, &sample->cylinderpressure) ||
+		MATCH("ref.switchmix", uddf_gasswitch, sample) ||
 		0;
 }
 
@@ -643,11 +782,10 @@ static void eventtime(char *buffer, void *_duration)
 
 static void try_to_match_autogroup(const char *name, char *buf)
 {
-	int len = strlen(name);
 	int autogroupvalue;
 
 	start_match("autogroup", name, buf);
-	if (MATCH(".autogroup.state", get_index, &autogroupvalue)) {
+	if (MATCH("state.autogroup", get_index, &autogroupvalue)) {
 		set_autogroup(autogroupvalue);
 		return;
 	}
@@ -656,18 +794,16 @@ static void try_to_match_autogroup(const char *name, char *buf)
 
 static void try_to_fill_dc_settings(const char *name, char *buf)
 {
-	int len = strlen(name);
-
 	start_match("divecomputerid", name, buf);
-	if (MATCH("divecomputerid.model", utf8_string, &cur_settings.dc.model))
+	if (MATCH("model.divecomputerid", utf8_string, &cur_settings.dc.model))
 		return;
-	if (MATCH("divecomputerid.deviceid", hex_value, &cur_settings.dc.deviceid))
+	if (MATCH("deviceid.divecomputerid", hex_value, &cur_settings.dc.deviceid))
 		return;
-	if (MATCH("divecomputerid.nickname", utf8_string, &cur_settings.dc.nickname))
+	if (MATCH("nickname.divecomputerid", utf8_string, &cur_settings.dc.nickname))
 		return;
-	if (MATCH("divecomputerid.serial", utf8_string, &cur_settings.dc.serial_nr))
+	if (MATCH("serial.divecomputerid", utf8_string, &cur_settings.dc.serial_nr))
 		return;
-	if (MATCH("divecomputerid.firmware", utf8_string, &cur_settings.dc.firmware))
+	if (MATCH("firmware.divecomputerid", utf8_string, &cur_settings.dc.firmware))
 		return;
 
 	nonmatch("divecomputerid", name, buf);
@@ -675,53 +811,51 @@ static void try_to_fill_dc_settings(const char *name, char *buf)
 
 static void try_to_fill_event(const char *name, char *buf)
 {
-	int len = strlen(name);
-
 	start_match("event", name, buf);
-	if (MATCH(".event", utf8_string, &cur_event.name))
+	if (MATCH("event", utf8_string, &cur_event.name))
 		return;
-	if (MATCH(".name", utf8_string, &cur_event.name))
+	if (MATCH("name", utf8_string, &cur_event.name))
 		return;
-	if (MATCH(".time", eventtime, &cur_event.time))
+	if (MATCH("time", eventtime, &cur_event.time))
 		return;
-	if (MATCH(".type", get_index, &cur_event.type))
+	if (MATCH("type", get_index, &cur_event.type))
 		return;
-	if (MATCH(".flags", get_index, &cur_event.flags))
+	if (MATCH("flags", get_index, &cur_event.flags))
 		return;
-	if (MATCH(".value", get_index, &cur_event.value))
+	if (MATCH("value", get_index, &cur_event.value))
 		return;
 	nonmatch("event", name, buf);
 }
 
-static int match_dc_data_fields(struct divecomputer *dc, const char *name, int len, char *buf)
+static int match_dc_data_fields(struct divecomputer *dc, const char *name, char *buf)
 {
-	if (MATCH(".maxdepth", depth, &dc->maxdepth))
+	if (MATCH("maxdepth", depth, &dc->maxdepth))
 		return 1;
-	if (MATCH(".meandepth", depth, &dc->meandepth))
+	if (MATCH("meandepth", depth, &dc->meandepth))
 		return 1;
-	if (MATCH(".depth.max", depth, &dc->maxdepth))
+	if (MATCH("max.depth", depth, &dc->maxdepth))
 		return 1;
-	if (MATCH(".depth.mean", depth, &dc->meandepth))
+	if (MATCH("mean.depth", depth, &dc->meandepth))
 		return 1;
-	if (MATCH(".duration", duration, &dc->duration))
+	if (MATCH("duration", duration, &dc->duration))
 		return 1;
-	if (MATCH(".divetime", duration, &dc->duration))
+	if (MATCH("divetime", duration, &dc->duration))
 		return 1;
-	if (MATCH(".divetimesec", duration, &dc->duration))
+	if (MATCH("divetimesec", duration, &dc->duration))
 		return 1;
-	if (MATCH(".surfacetime", duration, &dc->surfacetime))
+	if (MATCH("surfacetime", duration, &dc->surfacetime))
 		return 1;
-	if (MATCH(".airtemp", temperature, &dc->airtemp))
+	if (MATCH("airtemp", temperature, &dc->airtemp))
 		return 1;
-	if (MATCH(".watertemp", temperature, &dc->watertemp))
+	if (MATCH("watertemp", temperature, &dc->watertemp))
 		return 1;
-	if (MATCH(".temperature.air", temperature, &dc->airtemp))
+	if (MATCH("air.temperature", temperature, &dc->airtemp))
 		return 1;
-	if (MATCH(".temperature.water", temperature, &dc->watertemp))
+	if (MATCH("water.temperature", temperature, &dc->watertemp))
 		return 1;
-	if (MATCH(".surface.pressure", pressure, &dc->surface_pressure))
+	if (MATCH("pressure.surface", pressure, &dc->surface_pressure))
 		return 1;
-	if (MATCH(".water.salinity", salinity, &dc->salinity))
+	if (MATCH("salinity.water", salinity, &dc->salinity))
 		return 1;
 
 	return 0;
@@ -730,22 +864,20 @@ static int match_dc_data_fields(struct divecomputer *dc, const char *name, int l
 /* We're in the top-level dive xml. Try to convert whatever value to a dive value */
 static void try_to_fill_dc(struct divecomputer *dc, const char *name, char *buf)
 {
-	int len = strlen(name);
-
 	start_match("divecomputer", name, buf);
 
-	if (MATCH(".date", divedate, &dc->when))
+	if (MATCH("date", divedate, &dc->when))
 		return;
-	if (MATCH(".time", divetime, &dc->when))
+	if (MATCH("time", divetime, &dc->when))
 		return;
-	if (MATCH(".model", utf8_string, &dc->model))
+	if (MATCH("model", utf8_string, &dc->model))
 		return;
-	if (MATCH(".deviceid", hex_value, &dc->deviceid))
+	if (MATCH("deviceid", hex_value, &dc->deviceid))
 		return;
-	if (MATCH(".diveid", hex_value, &dc->diveid))
+	if (MATCH("diveid", hex_value, &dc->diveid))
 		return;
 
-	if (match_dc_data_fields(dc, name, len, buf))
+	if (match_dc_data_fields(dc, name, buf))
 		return;
 
 	nonmatch("divecomputer", name, buf);
@@ -754,17 +886,15 @@ static void try_to_fill_dc(struct divecomputer *dc, const char *name, char *buf)
 void add_gas_switch_event(struct dive *dive, struct divecomputer *dc, int seconds, int idx)
 {
 	/* The gas switch event format is insane. It will be fixed, I think */
-	int o2 = dive->cylinder[idx].gasmix.o2.permille;
-	int he = dive->cylinder[idx].gasmix.he.permille;
+	int o2 = get_o2(&dive->cylinder[idx].gasmix);
+	int he = get_he(&dive->cylinder[idx].gasmix);
 	int value;
 
-	if (!o2)
-		o2 = O2_IN_AIR;
 	o2 = (o2+5) / 10;
 	he = (he+5) / 10;
 	value = o2 + (he << 16);
 
-	add_event(dc, seconds, 11, 0, value, "gaschange");
+	add_event(dc, seconds, 25, 0, value, "gaschange");  /* SAMPLE_EVENT_GASCHANGE2 */
 }
 
 static void get_cylinderindex(char *buffer, void *_i)
@@ -787,51 +917,50 @@ static void get_sensor(char *buffer, void *_i)
 /* We're in samples - try to convert the random xml value to something useful */
 static void try_to_fill_sample(struct sample *sample, const char *name, char *buf)
 {
-	int len = strlen(name);
 	int in_deco;
 
 	start_match("sample", name, buf);
-	if (MATCH(".sample.pressure", pressure, &sample->cylinderpressure))
+	if (MATCH("pressure.sample", pressure, &sample->cylinderpressure))
 		return;
-	if (MATCH(".sample.cylpress", pressure, &sample->cylinderpressure))
+	if (MATCH("cylpress.sample", pressure, &sample->cylinderpressure))
 		return;
-	if (MATCH(".sample.cylinderindex", get_cylinderindex, &sample->sensor))
+	if (MATCH("cylinderindex.sample", get_cylinderindex, &sample->sensor))
 		return;
-	if (MATCH(".sample.sensor", get_sensor, &sample->sensor))
+	if (MATCH("sensor.sample", get_sensor, &sample->sensor))
 		return;
-	if (MATCH(".sample.depth", depth, &sample->depth))
+	if (MATCH("depth.sample", depth, &sample->depth))
 		return;
-	if (MATCH(".sample.temp", temperature, &sample->temperature))
+	if (MATCH("temp.sample", temperature, &sample->temperature))
 		return;
-	if (MATCH(".sample.temperature", temperature, &sample->temperature))
+	if (MATCH("temperature.sample", temperature, &sample->temperature))
 		return;
-	if (MATCH(".sample.sampletime", sampletime, &sample->time))
+	if (MATCH("sampletime.sample", sampletime, &sample->time))
 		return;
-	if (MATCH(".sample.time", sampletime, &sample->time))
+	if (MATCH("time.sample", sampletime, &sample->time))
 		return;
-	if (MATCH(".sample.ndl", sampletime, &sample->ndl))
+	if (MATCH("ndl.sample", sampletime, &sample->ndl))
 		return;
-	if (MATCH(".sample.in_deco", get_index, &in_deco)) {
+	if (MATCH("in_deco.sample", get_index, &in_deco)) {
 		sample->in_deco = (in_deco == 1);
 		return;
 	}
-	if (MATCH(".sample.stoptime", sampletime, &sample->stoptime))
+	if (MATCH("stoptime.sample", sampletime, &sample->stoptime))
 		return;
-	if (MATCH(".sample.stopdepth", depth, &sample->stopdepth))
+	if (MATCH("stopdepth.sample", depth, &sample->stopdepth))
 		return;
-	if (MATCH(".sample.cns", get_index, &sample->cns))
+	if (MATCH("cns.sample", get_index, &sample->cns))
 		return;
-	if (MATCH(".sample.po2", double_to_permil, &sample->po2))
+	if (MATCH("po2.sample", double_to_permil, &sample->po2))
 		return;
 
 	switch (import_source) {
 	case DIVINGLOG:
-		if (divinglog_fill_sample(sample, name, len, buf))
+		if (divinglog_fill_sample(sample, name, buf))
 			return;
 		break;
 
 	case UDDF:
-		if (uddf_fill_sample(sample, name, len, buf))
+		if (uddf_fill_sample(sample, name, buf))
 			return;
 		break;
 
@@ -866,23 +995,23 @@ static void divinglog_place(char *place, void *_location)
 	country = NULL;
 }
 
-static int divinglog_dive_match(struct dive *dive, const char *name, int len, char *buf)
+static int divinglog_dive_match(struct dive *dive, const char *name, char *buf)
 {
-	return	MATCH(".divedate", divedate, &dive->when) ||
-		MATCH(".entrytime", divetime, &dive->when) ||
-		MATCH(".divetime", duration, &dive->dc.duration) ||
-		MATCH(".depth", depth, &dive->dc.maxdepth) ||
-		MATCH(".depthavg", depth, &dive->dc.meandepth) ||
-		MATCH(".tanktype", utf8_string, &dive->cylinder[0].type.description) ||
-		MATCH(".tanksize", cylindersize, &dive->cylinder[0].type.size) ||
-		MATCH(".presw", pressure, &dive->cylinder[0].type.workingpressure) ||
-		MATCH(".press", pressure, &dive->cylinder[0].start) ||
-		MATCH(".prese", pressure, &dive->cylinder[0].end) ||
-		MATCH(".comments", utf8_string, &dive->notes) ||
-		MATCH(".buddy.names", utf8_string, &dive->buddy) ||
-		MATCH(".country.name", utf8_string, &country) ||
-		MATCH(".city.name", utf8_string, &city) ||
-		MATCH(".place.name", divinglog_place, &dive->location) ||
+	return	MATCH("divedate", divedate, &dive->when) ||
+		MATCH("entrytime", divetime, &dive->when) ||
+		MATCH("divetime", duration, &dive->dc.duration) ||
+		MATCH("depth", depth, &dive->dc.maxdepth) ||
+		MATCH("depthavg", depth, &dive->dc.meandepth) ||
+		MATCH("tanktype", utf8_string, &dive->cylinder[0].type.description) ||
+		MATCH("tanksize", cylindersize, &dive->cylinder[0].type.size) ||
+		MATCH("presw", pressure, &dive->cylinder[0].type.workingpressure) ||
+		MATCH("press", pressure, &dive->cylinder[0].start) ||
+		MATCH("prese", pressure, &dive->cylinder[0].end) ||
+		MATCH("comments", utf8_string, &dive->notes) ||
+		MATCH("names.buddy", utf8_string, &dive->buddy) ||
+		MATCH("name.country", utf8_string, &country) ||
+		MATCH("name.city", utf8_string, &city) ||
+		MATCH("name.place", divinglog_place, &dive->location) ||
 		0;
 }
 
@@ -940,16 +1069,16 @@ uddf_datedata(mday, 0)
 uddf_datedata(hour, 0)
 uddf_datedata(min, 0)
 
-static int uddf_dive_match(struct dive *dive, const char *name, int len, char *buf)
+static int uddf_dive_match(struct dive *dive, const char *name, char *buf)
 {
-	return	MATCH(".datetime", uddf_datetime, &dive->when) ||
-		MATCH(".diveduration", duration, &dive->dc.duration) ||
-		MATCH(".greatestdepth", depth, &dive->dc.maxdepth) ||
-		MATCH(".date.year", uddf_year, &dive->when) ||
-		MATCH(".date.month", uddf_mon, &dive->when) ||
-		MATCH(".date.day", uddf_mday, &dive->when) ||
-		MATCH(".time.hour", uddf_hour, &dive->when) ||
-		MATCH(".time.minute", uddf_min, &dive->when) ||
+	return	MATCH("datetime", uddf_datetime, &dive->when) ||
+		MATCH("diveduration", duration, &dive->dc.duration) ||
+		MATCH("greatestdepth", depth, &dive->dc.maxdepth) ||
+		MATCH("year.date", uddf_year, &dive->when) ||
+		MATCH("month.date", uddf_mon, &dive->when) ||
+		MATCH("day.date", uddf_mday, &dive->when) ||
+		MATCH("hour.time", uddf_hour, &dive->when) ||
+		MATCH("minute.time", uddf_min, &dive->when) ||
 		0;
 }
 
@@ -963,7 +1092,7 @@ static degrees_t parse_degrees(char *buf, char **end)
 	int sign = 1, decimals = 6, value = 0;
 	degrees_t ret;
 
-	while (g_ascii_isspace(*buf))
+	while (isspace(*buf))
 		buf++;
 	switch (*buf) {
 	case '-':
@@ -1029,18 +1158,16 @@ static void gps_location(char *buffer, void *_dive)
 /* We're in the top-level dive xml. Try to convert whatever value to a dive value */
 static void try_to_fill_dive(struct dive *dive, const char *name, char *buf)
 {
-	int len = strlen(name);
-
 	start_match("dive", name, buf);
 
 	switch (import_source) {
 	case DIVINGLOG:
-		if (divinglog_dive_match(dive, name, len, buf))
+		if (divinglog_dive_match(dive, name, buf))
 			return;
 		break;
 
 	case UDDF:
-		if (uddf_dive_match(dive, name, len, buf))
+		if (uddf_dive_match(dive, name, buf))
 			return;
 		break;
 
@@ -1048,88 +1175,88 @@ static void try_to_fill_dive(struct dive *dive, const char *name, char *buf)
 		break;
 	}
 
-	if (MATCH(".number", get_index, &dive->number))
+	if (MATCH("number", get_index, &dive->number))
 		return;
-	if (MATCH(".tags", divetags, &dive->dive_tags))
+	if (MATCH("tags", divetags, dive->tag_list))
 		return;
-	if (MATCH(".tripflag", get_tripflag, &dive->tripflag))
+	if (MATCH("tripflag", get_tripflag, &dive->tripflag))
 		return;
-	if (MATCH(".date", divedate, &dive->when))
+	if (MATCH("date", divedate, &dive->when))
 		return;
-	if (MATCH(".time", divetime, &dive->when))
+	if (MATCH("time", divetime, &dive->when))
 		return;
-	if (MATCH(".datetime", divedatetime, &dive->when))
+	if (MATCH("datetime", divedatetime, &dive->when))
 		return;
 	/*
 	 * Legacy format note: per-dive depths and duration get saved
 	 * in the first dive computer entry
 	 */
-	if (match_dc_data_fields(&dive->dc, name, len, buf))
+	if (match_dc_data_fields(&dive->dc, name, buf))
 		return;
 
-	if (MATCH(".cylinderstartpressure", pressure, &dive->cylinder[0].start))
+	if (MATCH("cylinderstartpressure", pressure, &dive->cylinder[0].start))
 		return;
-	if (MATCH(".cylinderendpressure", pressure, &dive->cylinder[0].end))
+	if (MATCH("cylinderendpressure", pressure, &dive->cylinder[0].end))
 		return;
-	if (MATCH(".gps", gps_location, dive))
+	if (MATCH("gps", gps_location, dive))
 		return;
-	if (MATCH(".Place", gps_location, dive))
+	if (MATCH("Place", gps_location, dive))
 		return;
-	if (MATCH(".latitude", gps_lat, dive))
+	if (MATCH("latitude", gps_lat, dive))
 		return;
-	if (MATCH(".sitelat", gps_lat, dive))
+	if (MATCH("sitelat", gps_lat, dive))
 		return;
-	if (MATCH(".lat", gps_lat, dive))
+	if (MATCH("lat", gps_lat, dive))
 		return;
-	if (MATCH(".longitude", gps_long, dive))
+	if (MATCH("longitude", gps_long, dive))
 		return;
-	if (MATCH(".sitelon", gps_long, dive))
+	if (MATCH("sitelon", gps_long, dive))
 		return;
-	if (MATCH(".lon", gps_long, dive))
+	if (MATCH("lon", gps_long, dive))
 		return;
-	if (MATCH(".location", utf8_string, &dive->location))
+	if (MATCH("location", utf8_string, &dive->location))
 		return;
-	if (MATCH("dive.name", utf8_string, &dive->location))
+	if (MATCH("name.dive", utf8_string, &dive->location))
 		return;
-	if (MATCH(".suit", utf8_string, &dive->suit))
+	if (MATCH("suit", utf8_string, &dive->suit))
 		return;
-	if (MATCH(".divesuit", utf8_string, &dive->suit))
+	if (MATCH("divesuit", utf8_string, &dive->suit))
 		return;
-	if (MATCH(".notes", utf8_string, &dive->notes))
+	if (MATCH("notes", utf8_string, &dive->notes))
 		return;
-	if (MATCH(".divemaster", utf8_string, &dive->divemaster))
+	if (MATCH("divemaster", utf8_string, &dive->divemaster))
 		return;
-	if (MATCH(".buddy", utf8_string, &dive->buddy))
+	if (MATCH("buddy", utf8_string, &dive->buddy))
 		return;
-	if (MATCH("dive.rating", get_rating, &dive->rating))
+	if (MATCH("rating.dive", get_rating, &dive->rating))
 		return;
-	if (MATCH("dive.visibility", get_rating, &dive->visibility))
+	if (MATCH("visibility.dive", get_rating, &dive->visibility))
 		return;
-	if (MATCH(".cylinder.size", cylindersize, &dive->cylinder[cur_cylinder_index].type.size))
+	if (MATCH("size.cylinder", cylindersize, &dive->cylinder[cur_cylinder_index].type.size))
 		return;
-	if (MATCH(".cylinder.workpressure", pressure, &dive->cylinder[cur_cylinder_index].type.workingpressure))
+	if (MATCH("workpressure.cylinder", pressure, &dive->cylinder[cur_cylinder_index].type.workingpressure))
 		return;
-	if (MATCH(".cylinder.description", utf8_string, &dive->cylinder[cur_cylinder_index].type.description))
+	if (MATCH("description.cylinder", utf8_string, &dive->cylinder[cur_cylinder_index].type.description))
 		return;
-	if (MATCH(".cylinder.start", pressure, &dive->cylinder[cur_cylinder_index].start))
+	if (MATCH("start.cylinder", pressure, &dive->cylinder[cur_cylinder_index].start))
 		return;
-	if (MATCH(".cylinder.end", pressure, &dive->cylinder[cur_cylinder_index].end))
+	if (MATCH("end.cylinder", pressure, &dive->cylinder[cur_cylinder_index].end))
 		return;
-	if (MATCH(".weightsystem.description", utf8_string, &dive->weightsystem[cur_ws_index].description))
+	if (MATCH("description.weightsystem", utf8_string, &dive->weightsystem[cur_ws_index].description))
 		return;
-	if (MATCH(".weightsystem.weight", weight, &dive->weightsystem[cur_ws_index].weight))
+	if (MATCH("weight.weightsystem", weight, &dive->weightsystem[cur_ws_index].weight))
 		return;
 	if (MATCH("weight", weight, &dive->weightsystem[cur_ws_index].weight))
 		return;
-	if (MATCH(".o2", gasmix, &dive->cylinder[cur_cylinder_index].gasmix.o2))
+	if (MATCH("o2", gasmix, &dive->cylinder[cur_cylinder_index].gasmix.o2))
 		return;
-	if (MATCH(".o2percent", gasmix, &dive->cylinder[cur_cylinder_index].gasmix.o2))
+	if (MATCH("o2percent", gasmix, &dive->cylinder[cur_cylinder_index].gasmix.o2))
 		return;
-	if (MATCH(".n2", gasmix_nitrogen, &dive->cylinder[cur_cylinder_index].gasmix))
+	if (MATCH("n2", gasmix_nitrogen, &dive->cylinder[cur_cylinder_index].gasmix))
 		return;
-	if (MATCH(".he", gasmix, &dive->cylinder[cur_cylinder_index].gasmix.he))
+	if (MATCH("he", gasmix, &dive->cylinder[cur_cylinder_index].gasmix.he))
 		return;
-	if (MATCH(".divetemperature.air", temperature, &dive->airtemp))
+	if (MATCH("air.divetemperature", temperature, &dive->airtemp))
 		return;
 
 	nonmatch("dive", name, buf);
@@ -1138,19 +1265,17 @@ static void try_to_fill_dive(struct dive *dive, const char *name, char *buf)
 /* We're in the top-level trip xml. Try to convert whatever value to a trip value */
 static void try_to_fill_trip(dive_trip_t **dive_trip_p, const char *name, char *buf)
 {
-	int len = strlen(name);
-
 	start_match("trip", name, buf);
 
 	dive_trip_t *dive_trip = *dive_trip_p;
 
-	if (MATCH(".date", divedate, &dive_trip->when))
+	if (MATCH("date", divedate, &dive_trip->when))
 		return;
-	if (MATCH(".time", divetime, &dive_trip->when))
+	if (MATCH("time", divetime, &dive_trip->when))
 		return;
-	if (MATCH(".location", utf8_string, &dive_trip->location))
+	if (MATCH("location", utf8_string, &dive_trip->location))
 		return;
-	if (MATCH(".notes", utf8_string, &dive_trip->notes))
+	if (MATCH("notes", utf8_string, &dive_trip->notes))
 		return;
 
 	nonmatch("trip", name, buf);
@@ -1168,7 +1293,7 @@ static void try_to_fill_trip(dive_trip_t **dive_trip_p, const char *name, char *
  * to make a dive valid, but if it has no location, no date and no
  * samples I'm pretty sure it's useless.
  */
-static gboolean is_dive(void)
+static bool is_dive(void)
 {
 	return (cur_dive &&
 		(cur_dive->location || cur_dive->when || cur_dive->dc.samples));
@@ -1210,17 +1335,8 @@ static void dc_settings_start(void)
 
 static void dc_settings_end(void)
 {
-	struct device_info *info;
-
-	info = create_device_info(cur_settings.dc.model, cur_settings.dc.deviceid);
-	if (info) {
-		if (!info->serial_nr && cur_settings.dc.serial_nr)
-			info->serial_nr = strdup(cur_settings.dc.serial_nr);
-		if (!info->firmware && cur_settings.dc.firmware)
-			info->firmware = strdup(cur_settings.dc.firmware);
-		if (!info->nickname && cur_settings.dc.nickname)
-			info->nickname = strdup(cur_settings.dc.nickname);
-	}
+	create_device_node(cur_settings.dc.model, cur_settings.dc.deviceid, cur_settings.dc.serial_nr,
+			   cur_settings.dc.firmware, cur_settings.dc.nickname);
 	reset_dc_settings();
 }
 
@@ -1393,47 +1509,52 @@ static void entry(const char *name, char *buf)
 
 static const char *nodename(xmlNode *node, char *buf, int len)
 {
+	int levels = 2;
+	char *p = buf;
+
 	if (!node || !node->name)
 		return "root";
 
-	buf += len;
-	*--buf = 0;
-	len--;
+	if (node->parent && !strcmp(node->name, "text"))
+		node = node->parent;
+
+	/* Make sure it's always NUL-terminated */
+	p[--len] = 0;
 
 	for(;;) {
 		const char *name = node->name;
-		int i = strlen(name);
-		while (--i >= 0) {
-			unsigned char c = name[i];
-			*--buf = tolower(c);
+		char c;
+		while ((c = *name++) != 0) {
+			/* Cheaper 'tolower()' for ASCII */
+			c = (c >= 'A' && c <= 'Z') ? c - 'A' + 'a' : c;
+			*p++ = c;
 			if (!--len)
 				return buf;
 		}
+		*p = 0;
 		node = node->parent;
 		if (!node || !node->name)
 			return buf;
-		*--buf = '.';
+		*p++ = '.';
 		if (!--len)
 			return buf;
+		if (!--levels)
+			return buf;
 	}
 }
 
-#define MAXNAME 64
+#define MAXNAME 32
 
 static void visit_one_node(xmlNode *node)
 {
 	char *content;
-	char buffer[MAXNAME];
+	static char buffer[MAXNAME];
 	const char *name;
 
 	content = node->content;
 	if (!content || xmlIsBlankNode(node))
 		return;
 
-	/* Don't print out the node name if it is "text" */
-	while (!node->name || !strcmp(node->name, "text"))
-		node = node->parent;
-
 	name = nodename(node, buffer, sizeof(buffer));
 
 	entry(name, content);
@@ -1579,7 +1700,7 @@ const char *preprocess_divelog_de(const char *buffer)
 }
 
 void parse_xml_buffer(const char *url, const char *buffer, int size,
-			struct dive_table *table, GError **error)
+			struct dive_table *table, const char **params, char **error)
 {
 	xmlDoc *doc;
 	const char *res = preprocess_divelog_de(buffer);
@@ -1590,13 +1711,13 @@ void parse_xml_buffer(const char *url, const char *buffer, int size,
 		free((char *)res);
 
 	if (!doc) {
-		fprintf(stderr, _("Failed to parse '%s'.\n"), url);
-		parser_error(error, _("Failed to parse '%s'"), url);
+		fprintf(stderr, translate("gettextFromC","Failed to parse '%s'.\n"), url);
+		parser_error(error, translate("gettextFromC","Failed to parse '%s'"), url);
 		return;
 	}
 	reset_all();
 	dive_start();
-	doc = test_xslt_transforms(doc, error);
+	doc = test_xslt_transforms(doc, params, error);
 	traverse(xmlDocGetRootElement(doc));
 	dive_end();
 	xmlFreeDoc(doc);
@@ -1632,6 +1753,10 @@ extern int dm4_events(void *handle, int columns, char **data, char **column)
 				/* 6 Mandatory safety stop ceiling error */
 				cur_event.name = strdup("violation");
 				break;
+			case 7:
+				/* 7 Below deco floor */
+				cur_event.name = strdup("below floor");
+				break;
 			case 8:
 				/* 8 Dive time alarm */
 				cur_event.name = strdup("divetime");
@@ -1650,6 +1775,10 @@ extern int dm4_events(void *handle, int columns, char **data, char **column)
 				/* 12 High ppO2 */
 				cur_event.name = strdup("PO2");
 				break;
+			case 13:
+				/* 13 Air time */
+				cur_event.name = strdup("airtime");
+				break;
 			case 18:
 				/* 18 Ceiling error */
 				cur_event.name = strdup("ceiling");
@@ -1730,9 +1859,9 @@ extern int dm4_dive(void *param, int columns, char **data, char **column)
 	if (data[6])
 		cur_dive->dc.maxdepth.mm = atof(data[6]) * 1000;
 	if (data[8])
-		cur_dive->dc.airtemp.mkelvin = (atoi(data[8]) + 273.15) * 1000;
+		cur_dive->dc.airtemp.mkelvin = C_to_mkelvin(atoi(data[8]));
 	if (data[9])
-		cur_dive->dc.watertemp.mkelvin  = (atoi(data[9]) + 273.15) * 1000;
+		cur_dive->dc.watertemp.mkelvin  = C_to_mkelvin(atoi(data[9]));
 
 	/*
 	 * TODO: handle multiple cylinders
@@ -1772,7 +1901,7 @@ extern int dm4_dive(void *param, int columns, char **data, char **column)
 			cur_sample->depth.mm = cur_dive->dc.maxdepth.mm;
 
 		if (tempBlob)
-			cur_sample->temperature.mkelvin = (tempBlob[i] + 273.15) * 1000;
+			cur_sample->temperature.mkelvin = C_to_mkelvin(tempBlob[i]);
 		if (data[19] && data[19][0])
 			cur_sample->cylinderpressure.mbar = pressureBlob[i] ;
 		sample_end();
@@ -1781,7 +1910,7 @@ extern int dm4_dive(void *param, int columns, char **data, char **column)
 	snprintf(get_events, sizeof(get_events) - 1, get_events_template, cur_dive->number);
 	retval = sqlite3_exec(handle, get_events, &dm4_events, 0, &err);
 	if (retval != SQLITE_OK) {
-		fprintf(stderr, _("Database query get_events failed.\n"));
+		fprintf(stderr, "%s", translate("gettextFromC","Database query get_events failed.\n"));
 		return 1;
 	}
 
@@ -1802,7 +1931,7 @@ extern int dm4_dive(void *param, int columns, char **data, char **column)
 }
 
 int parse_dm4_buffer(const char *url, const char *buffer, int size,
-			struct dive_table *table, GError **error)
+			struct dive_table *table, char **error)
 {
 	int retval;
 	char *err = NULL;
@@ -1816,14 +1945,14 @@ int parse_dm4_buffer(const char *url, const char *buffer, int size,
 	retval = sqlite3_open(url,&handle);
 
 	if(retval) {
-		fprintf(stderr, _("Database connection failed '%s'.\n"), url);
+		fprintf(stderr, translate("gettextFromC","Database connection failed '%s'.\n"), url);
 		return 1;
 	}
 
 	retval = sqlite3_exec(handle, get_dives, &dm4_dive, handle, &err);
 
 	if (retval != SQLITE_OK) {
-		fprintf(stderr, _("Database query failed '%s'.\n"), url);
+		fprintf(stderr, translate("gettextFromC","Database query failed '%s'.\n"), url);
 		return 1;
 	}
 
@@ -1841,9 +1970,6 @@ void parse_xml_exit(void)
 	xmlCleanupParser();
 }
 
-/* Maybe we'll want a environment variable that can override this.. */
-static const char *xslt_path = XSLT ":xslt:.";
-
 static xsltStylesheetPtr try_get_stylesheet(const char *path, int len, const char *name)
 {
 	xsltStylesheetPtr ret;
@@ -1854,7 +1980,11 @@ static xsltStylesheetPtr try_get_stylesheet(const char *path, int len, const cha
 		return NULL;
 
 	memcpy(filename, path, len);
-	filename[len] = G_DIR_SEPARATOR;
+#ifdef WIN32
+	filename[len] = '\\';
+#else
+	filename[len] = '/';
+#endif
 	memcpy(filename + len + 1, name, namelen+1);
 
 	ret = NULL;
@@ -1904,10 +2034,11 @@ static struct xslt_files {
 	{ "UDDF", "uddf.xslt" },
 	{ "profile", "udcf.xslt" },
 	{ "Divinglog", "DivingLog.xslt" },
+	{ "csv", "csv2xml.xslt" },
 	{ NULL, }
 };
 
-static xmlDoc *test_xslt_transforms(xmlDoc *doc, GError **error)
+static xmlDoc *test_xslt_transforms(xmlDoc *doc, const char **params, char **error)
 {
 	struct xslt_files *info = xslt_files;
 	xmlDoc *transformed;
@@ -1932,12 +2063,13 @@ static xmlDoc *test_xslt_transforms(xmlDoc *doc, GError **error)
 		xmlSubstituteEntitiesDefault(1);
 		xslt = get_stylesheet(info->file);
 		if (xslt == NULL) {
-			parser_error(error, _("Can't open stylesheet (%s)/%s"), xslt_path, info->file);
+			parser_error(error, translate("gettextFromC","Can't open stylesheet (%s)/%s"), xslt_path, info->file);
 			return doc;
 		}
-		transformed = xsltApplyStylesheet(xslt, doc, NULL);
+		transformed = xsltApplyStylesheet(xslt, doc, params);
 		xmlFreeDoc(doc);
 		xsltFreeStylesheet(xslt);
+
 		return transformed;
 	}
 	return doc;
